//
//    Tracker produces a constantly updating display of the amplitude
//    and phase of line features in the spectrum of a channel. 
//    The configuration for Tracker is specified by an ascii file 
//    passed to Tracker() as an argument. The file contains a single
//    line:
//
//    Line 1: <channel> <tAverage> <frequency>
//
//    The fields have the following meaning:
//    <channel>      The name of the channel from which the data are derived.
//    <tAverage>     The averaging time over which the spectrum is taken.
//    <frequency>    The approximate frequency of the line feature to be tracked.
//    
//
//

#include "DTracker.hh"

#ifndef __CINT__
Dacc In("/relativity/LIGO1/H-609940/*.F");
#endif

//--------------------------------------  Channel descriptor class
class Chan {
public:
  Chan() {}
  ~Chan () {}
  char mName[40];
  double mRate;
};


class DTracker : public DatEnv {
  public:
  DTracker(int argc, const char* argv[]);
  ~DTracker();
  void ProcessData(void);
  private:
  Chan chanv;
  int nWin;  // number of taper windows to use
  int nPoint;
  int paddingFactor;
  int paddedLength;
  int maxLines;  // maximum number of lines to remove
  int nFreq;
  int nRemoved;
  int nDisplay;
  int nBins;     // number of (padded) frequency bins (either side) to search near fpreset
  int maxPass;   // the number of passes to make within the line removal algorithm 
  int nPass;
  float fpreset;
  float nwdt;    // time-freq bandwidth 
  float *data;
  float *initial_data;
  float *mtap_spec_init;
  float *mtap_spec_final;
  float displaytime;
  float t1;
  float *amp;
  float dbin;
  float ffit;
  float intercept;
  float slope;
  float deviation;
  float maxamp;
  float displayamp;
  // int code;
  double tsecs;  // sample time in seconds
  bool estimate;     // set true if you want to estimate the best-fit frequency 
  bool first;
  int top,nPlot,index;
  int minbin,maxbin;
  float creal,cimag;
  float *phase,amp1,phase1,phase2,*times,*linfitx,*linfity;
  float mOffset;
  float binpreset;
  float *plotAmp;
  float *plotPhase;
  double time;   // time from beginning or from start of new locked section
  RemovedLines *line_list;
  TCanvas *c;
  TPad *ampPad;
  TPad *phasePad;
  TH1F *ampGraph;
  TH1F *phaseGraph;
  const float offScaleAmp;
  const float offScalePhase;
  const float fpi;
  const float twopi;
};     

using namespace std;

void
Tracker(const char* config) {
    DTracker xxx(1, &config);
    xxx.MainLoop();
}

//--------------------------------------  Define the main program in background
#ifndef __CINT__
// equivalent to EXEROOT - see DatEnv.hh
TROOT root("DatEnv", "Graphical background monitor"); 
int main(int argc, const char* argv[]) { 
    TApplication theApp("ACO Multitaper Line Tracker", &argc, const_cast<char**>(argv)); 
    DTracker MyObject(argc,argv); 
    MyObject.MainLoop();
    return 0;}   
#endif


DTracker::DTracker(int argc, const char* argv[]) 
  :DatEnv(argc, argv), nWin(5), paddingFactor(4), maxLines(64), nDisplay(15), nBins(15), 
  maxPass(1),  nwdt(3.0), estimate(false), offScaleAmp(-1.0), offScalePhase(-250), 
  fpi(M_PI), twopi(2.0*M_PI)
{
   const char* config;
   config = argv[argc-1];
   if (!config || !*config) {
     cerr<< "# Tracker: configuration file not specified" << endl;
     return;
   }

   //---------------------------------  Get parameters.
   ifstream ifs(config, ios::in);
   if (!ifs) {
     cerr << "# Tracker: unable to open configuration file " << config << endl;
     return;
   }

   cerr << "# Tracker: reading parameters from config file: " << config << endl;
   
   ifs >> chanv.mName >> tsecs >> fpreset; 
   cerr << "# Tracker: using channel " << chanv.mName << " with sample time = " 
	   << tsecs << " seconds" <<endl;
   cerr << "# Tracker: looking to track line near " << fpreset << " Hz" <<endl;
   ifs.close();

   mIn.addChannel(chanv.mName);
   mIn.setBuffer(2);    // Read upto 2 frames ahead
   mIn.setStride(tsecs);   // Amount of data to take up

   //initialise
   first=true;

   // initialise ROOT graphics
   gROOT->Reset();
   c = new TCanvas("c", "Multitaper Line Tracker",10,10,788,750);
   c->SetHighLightColor(2);
   c->SetFillColor(33); // Blue-grey in the default ROOT palette
   c->SetBorderSize(2);
   c->SetLeftMargin(0.12);
   c->SetRightMargin(0.0);
   c->SetTopMargin(0.08);
   c->SetBottomMargin(0.12);
   c->SetEditable();
   ampPad = new TPad("ampPad","Amplitude",0.025,0.49,0.975,0.92,21);
   ampPad->SetFillColor(10);
   ampPad->SetBorderMode(0);
   ampPad->SetGrid();
   ampPad->SetTicks();
   ampPad->SetTitle("Amplitude");
   ampPad->SetFrameFillColor(10);
   ampPad->Draw();
   phasePad = new TPad("phasePad","Phase",0.03,0.025,0.975,0.46,21);
   phasePad->SetFillColor(10);
   phasePad->SetBorderMode(0);
   phasePad->SetGrid();
   phasePad->SetTicks();
   phasePad->SetTitle("Phase");
   phasePad->SetFrameFillColor(10);
   phasePad->Draw();

   gStyle->SetOptStat(0);   //   Don't show statistics
   gROOT->SetInterrupt(kTRUE); 
   
}
 
void
DTracker::ProcessData() {    

  if (first) {
    char UTCdate[256];
    TimeStr((mIn.refData(chanv.mName))->getStartTime(), UTCdate,"%H:%N:%S %M %d %Y %Z");
    cerr << "# Tracker: Starting reading frame " << mIn.getFile() << endl;
    cerr << "\a# Tracker: New Locked Segment at UTC time " << UTCdate << endl;

    chanv.mRate=1./((mIn.refData(chanv.mName))->getTStep());  // sample rate
    cerr<< "# Tracker: Sample Rate = " << chanv.mRate << endl;
    nPoint=int(tsecs*chanv.mRate);  // total number of data points   
    paddedLength=paddingFactor*nPoint;   // number of points with zero pading
    nFreq=1+paddedLength/2;     // num frequencies including DC, Nyquist 

    // allocate arrays 
    data=new float[nPoint];
    clear(data,nPoint,1);    // zero the array
    initial_data=new float[nPoint];
    mtap_spec_init=new float[nFreq];
    mtap_spec_final=new float[nFreq];
    line_list=new RemovedLines[maxLines];

    amp=new float[nDisplay];	 
    phase=new float[nDisplay];
    plotAmp=new float[nDisplay];
    phase=new float[nDisplay];
    plotPhase=new float[nDisplay];
    times=new float[nDisplay];
    linfitx=new float[nDisplay];
    linfity=new float[nDisplay];

    for (int i=0;i<nDisplay;i++) {
	 plotAmp[i]=offScaleAmp;
	 plotPhase[i]=offScalePhase;
    }

    // First call initialisations
    top=0;
    displayamp=1.0;
    nPass=1;
    phase1=0.0;
    ffit=fpreset;
    t1=time=0.0;
    
    // Work out range of bins to look for line
    binpreset=fpreset*2.0*nFreq/chanv.mRate;
    minbin=(int) binpreset-nBins;
    if (minbin<0) minbin=0;
    maxbin=(int) binpreset+nBins;
    if (maxbin>nFreq) maxbin=nFreq;
    
    // Total time displayed
    displaytime=nDisplay*(nPoint/chanv.mRate);
    
    ampGraph=new TH1F("ampGraph","Amplitude",nDisplay,t1,t1+displaytime);
    phaseGraph=new TH1F("phaseGraph","Phase",nDisplay,t1,t1+displaytime);

    first=false;  
  }

  // copy data into float array data and
  // check we have the number of points we expect
  int ptsRetd=(mIn.refData(chanv.mName))->getData(nPoint,data);
  if (ptsRetd != nPoint) {
    cerr << "Tracker: Points requested = " << nPoint << 
	 ",  Points returned = " << ptsRetd <<endl;
    abort();
  }
  time+=tsecs;

  // remove the spectral lines from the data set 
  remove_spectral_lines(data,nPoint,paddedLength,nwdt,nWin,maxLines,maxPass,
				    &nRemoved,line_list,mtap_spec_init,mtap_spec_final,
				    0,minbin,maxbin);

  // if we fail to remove a line, amplitude set to zero, phase RETAINS PRIOR VALUE
  amp1=0.0;  
  // look in the list of removed lines for the right one 
  for (int i=0;i<nRemoved;i++) {
    // the closest bin to our estimated frequency 
    dbin=binpreset-line_list[i].index;
    if (fabs(dbin) <= nBins) {
	 creal=line_list[i].re+dbin*line_list[i].dcdbr+
	   0.5*dbin*dbin*line_list[i].d2cdb2r;
	 cimag=line_list[i].im+dbin*line_list[i].dcdbi+
	   0.5*dbin*dbin*line_list[i].d2cdb2i;
	 amp1=2.0*sqrt(creal*creal+cimag*cimag);
	 phase1=atan2(cimag,creal)+2.0*fpi*fmod(fpreset*time,1.0);
	 break;
    }
  }
  
  // save data in a circular buffers *[0..nDisplay-1]  
  amp[top]=amp1;
  phase[top]=fmod_2pi(phase1);
  times[top]=time;

  // how many values are we going to output to the graph? 
  nPlot=(nPass<nDisplay)?nPass:nDisplay;
  // cut out a piece for the linear fit 
  if (nPass>=2) {
    // adjust the phases to avoid boundary jumps 
    mOffset=0.0;
    index=mod_n(top-nPlot+1,nDisplay);
    linfitx[0]=times[index];
    linfity[0]=phase[index];
    for (int i=1;i<nPlot;i++) {
	 index=mod_n(top+i-nPlot+1,nDisplay);
	 linfitx[i]=times[index];   
	 if ( (phase[index]-phase[mod_n(index-1,nDisplay)]) > fpi) {
	   mOffset-=twopi; 
	 }
	 else if ( (phase[index]-phase[mod_n(index-1,nDisplay)]) < -fpi) {
	   mOffset+=twopi;
	 }
	 linfity[i]=phase[index]+mOffset;
    }
   
    // do a robust linear fit
    medfit(linfitx-1,linfity-1,nPlot,&intercept,&slope,&deviation);
    
    // now see what frequency the best fit corresponds to 
    ffit=fpreset-slope/twopi;
    
    // if we are assuming a fixed frequency (not adapting) 
    if (!estimate) {
	 slope=intercept=0.0;
    }

    // print out amplitude if non-zero 
    maxamp=0.0;
    for (int i=0;i<nPlot;i++) {
	 index=mod_n(top+i-nPlot+1,nDisplay);
	 if (amp[index]>0.0)
	   plotAmp[i]=amp[index];
	 else
	   plotAmp[i]=offScaleAmp; // won't appear on the graph - out of bounds
	 
	 if (amp[index]>maxamp) maxamp=amp[index];
    }

    //for (i=0;i<nDisplay;i++) {
    //	 cout << linfitx[i] << "\t" << plotAmp[i] << endl;
    //}

    // print out phase  if non-zero amplitude 
    for (int i=0;i<nPlot;i++) {
	 phase2=linfity[i];
	 phase2=fmod_2pi((phase2-slope*linfitx[i]-intercept));
	 if (phase2>fpi)
	   phase2-=twopi;
	 phase2=(180.0/fpi)*phase2;
	 index=mod_n(top+i-nPlot+1,nDisplay);
	 if (amp[index]>0.0)
	   plotPhase[i]=phase2; 
	 else
	   plotPhase[i]=offScalePhase; // won't appear on the graph - out of bounds 
    }

    //    for (i=0;i<nDisplay;i++) {
    //      cout << linfitx[i] << "\t" << plotPhase[i] << endl;
    //    }

    // set up scale of the x-axis 
    t1=linfitx[0];
    // set up scale of the amplitude graph y-axis 
    if (maxamp>0.9*displayamp) { 
	 displayamp=1.3*maxamp;
	 cerr << "\aTracker: Line at " << fpreset << " -- amplitude just increased\n" << endl;
	 cerr << "Displayamp now " << displayamp << endl;
    }
    else if (maxamp<0.4*displayamp && maxamp>0.0)
	 displayamp=1.3*maxamp;           
    
  }

  // now display

  c->cd();
  TPaveText *pt = new TPaveText(0.3, 0.935, 0.7, 0.99,"NDC");
  pt->SetFillColor(41);
  ostringstream ost;
  ost << "Best fit frequency: " << ffit << ends;
  pt->AddText(ost.str().c_str());
  pt->Draw();

  *ampGraph=TH1F("ampGraph","Amplitude",nDisplay,t1,t1+displaytime); 
  ampGraph->SetMinimum(0.0);
  ampGraph->SetMarkerColor(4);   // Blue in the default ROOT palette 
  ampGraph->SetMarkerStyle(8);
  ampGraph->SetMarkerSize(0.7);
  ampGraph->SetLineColor(4);     // Blue in the default ROOT palette
  ampGraph->SetXTitle("Time (s)");

  *phaseGraph=TH1F("phaseGraph","Phase",nDisplay,t1,t1+displaytime);
  phaseGraph->SetMaximum(180.0);
  phaseGraph->SetMinimum(-180.0);
  phaseGraph->SetMarkerColor(2);   // Red in the default ROOT palette
  phaseGraph->SetMarkerStyle(8);
  phaseGraph->SetMarkerSize(0.7);
  phaseGraph->SetLineColor(2);     // Red in the default ROOT palette
  phaseGraph->SetXTitle("Time (s)");
 
  for (int i=0;i<nDisplay;i++) {
	  ampGraph->Fill(linfitx[i],plotAmp[i]);
	  phaseGraph->Fill(linfitx[i],plotPhase[i]);
  }                     

  ampPad->cd();
  ampGraph->SetMaximum(displayamp);
  ampGraph->Draw("CP");       // Draw points
  ampPad->Modified();
  c->Update();

  phasePad->cd();
  phaseGraph->Draw("P");
  phasePad->Modified();
  c->Update();

  // get ready for next set 
  nPass++;
  top=mod_n(top+1,nDisplay);

  return;
}

DTracker::~DTracker() {
  //---------------------------------  Clean up at the end
  In.rmChannel(chanv.mName);

  cout << "Tracker finished" << endl;
}     

