/*
 * h(t) monitor - Generation of time-domain calibrated data.
 *
 * Jordi Burguet-Castell, Shourov K. Chatterji, Xavier Siemens.
 * Aug, Sep, Oct 2008.
 */

// $Id: hoftMon.cc,v 1.3 2008/11/04 18:02:50 jordi Exp $

#include "hoftMon.hh"


// The next three lines are needed if you are going to generate triggers.
// The text in PIDTITLE should describe the monitor function.
#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/hoftMon/hoftMon.cc,v 1.3 2008/11/04 18:02:50 jordi Exp $"
#define PIDTITLE  "Generate calibrated h(t)"
#include "ProcIdent.hh"

#include "Dacc.hh"
#include <iostream>
#include <string>
#include <cstdlib>

extern "C" {
#include <lal/ReadFiltersFile.h>   // new file with Xavi's function
#include <lal/AVFactories.h>       // to use XLALCreateREAL4Vector
}

using namespace std;


INT4 lalDebugLevel = 0;  /* for debugging calls to LAL functions, put it to 3 */

// Macro for log output.
#define LOG(LEVEL) if (mVerboseLevel >= LEVEL) cout << Now() << ": "

// Macro for common checks.
#define CHECK(COND) \
    do {								\
	if (COND != 0) {						\
	    cerr << "Error in " << __FILE__ << " at " << __LINE__ << endl; \
	    finish();  return;						\
	}								\
    } while (0)


// Generate main routine.
EXECDAT(hoftMon)


/*
 * Constructor.
 */
hoftMon::hoftMon(int argc, const char *argv[])
  : DatEnv(argc, argv), mStep(16.0),
    mVerboseLevel(0), mIteration(0)
{
    int err = 0;  /* error code */

    LOG(3) << "Initializing\n";
    mInputData.Cinv = NULL;       // filters file still not read
    mInputData.AS_Q.data = NULL;  // lal vectors not allocated yet
    err = initialize_input_data(argc, argv);
    CHECK(err);

    // Set the data accessor.
    getDacc().setStride(mStep);  // we'll read chunks of mStep seconds
    for (int i = 0; i < 4; i++)
	getDacc().addChannel(mChanNames[i]);  // add channels
}



/*
 * Destructor.
 */
hoftMon::~hoftMon() 
{
    if (mInputData.AS_Q.data != NULL)
	free_data();  // frees the LAL channels

    if (mInputData.Cinv != NULL)
        destroy_initialization();    // destroys filters file
    
    LOG(3) << "hoftMon is finished\n";
}


/*
 * Handle messages.
 */
void hoftMon::Attention(void)
{
    MonServer::Attention();
}



/* ------------------------- Main functions ---------------------------- */


/*
 * Process one data stride. This function gets called each time a new
 * frame is ready to be processed.
 */
void hoftMon::ProcessData()
{
    int err = 0;  /* error code */

    LOG(3) << "Starting ProcessData (iteration " << mIteration << ")\n";
    
    const TSeries* dmt_chans[4];
    err = get_channels(dmt_chans);    // fills dmt_chans with the DMT channels
    CHECK(err);

    if (mIteration == 0) {
	LOG(3) << "Allocating memory for the LAL channels\n";
	err = allocate_data(dmt_chans);  // allocate memory for the LAL channels
	CHECK(err);
    }

    LOG(3) << "Reading data: fill LAL input and output structures\n";
    read_data(dmt_chans);  // fills LAL input and output structures

    if (mIteration >= 2) {
	LOG(3) << "Calling ComputeStrain\n";
	LALStatus status;
	status.statusPtr = NULL;   // same thing as "static LALStatus status;"
	LALComputeStrain(&status, &mOutputData, &mInputData);

	err = write_frames(dmt_chans);
	CHECK(err);
    }

    mIteration++;
}






/*
 * Fills mInputData with the times (epoch and deltaT) and channels,
 * extracted from the DMT.
 *
 * To be called in ProcessData().
 */
void hoftMon::read_data(const TSeries* dmt_chans[])
{
    /* Get epoch, deltaT's and numer of samples for each channel */

    LIGOTimeGPS epoch;
    epoch.gpsSeconds     = (INT4) dmt_chans[0]->getStartTime().getS() - 32;
    epoch.gpsNanoSeconds = (INT4) dmt_chans[0]->getStartTime().getN();
    // The 32 seconds shift is done to get the epoch of two strides of data ago
    
    REAL8 deltaT[4];
    TSeries::size_type nsamples[4];
    for (int i = 0; i < 4; i++) {
	deltaT[i] = (REAL8) dmt_chans[i]->getTStep().GetSecs();
	nsamples[i] = dmt_chans[i]->getNSample();
    }

    REAL8 To = mInputData.To;


    /* Input channels, copied from the dmt channels */

    REAL4* chan_data[4] = {
	mInputData.AS_Q.data->data,
	mInputData.DARM.data->data,
	mInputData.DARM_ERR.data->data,
	mInputData.EXC.data->data
    };
    
    for (int i = 0; i < 4; i++) {
	int n = (int) nsamples[i];

	// Shift all data to the left.
	for (int j = 0; j < 2*n; j++)
	    chan_data[i][j] = chan_data[i][j+n];
	
	// Write new data in the last part of the array.
	dmt_chans[i]->getData(nsamples[i], chan_data[i] + 2 * nsamples[i]);
    }


    /* Fill times (epoch and deltaT) */

    mInputData.AS_Q.epoch        = epoch;
    mInputData.DARM.epoch        = epoch;
    mInputData.DARM_ERR.epoch    = epoch;
    mInputData.EXC.epoch         = epoch;
    mOutputData.h.epoch          = epoch;
    mOutputData.hC.epoch         = epoch;
    mOutputData.hR.epoch         = epoch;
    mOutputData.alpha.epoch      = epoch;
    mOutputData.beta.epoch       = epoch;
    mOutputData.alphabeta.epoch  = epoch;

    mInputData.AS_Q.deltaT       = deltaT[0];
    mInputData.DARM.deltaT       = deltaT[1];
    mInputData.DARM_ERR.deltaT   = deltaT[2];
    mInputData.EXC.deltaT        = deltaT[3];
    mOutputData.h.deltaT         = deltaT[0];  // same as AS_Q
    mOutputData.hC.deltaT        = deltaT[0];  // same as AS_Q
    mOutputData.hR.deltaT        = deltaT[0];  // same as AS_Q
    mOutputData.alpha.deltaT     = To;
    mOutputData.beta.deltaT      = To;
    mOutputData.alphabeta.deltaT = To;
}


#include <fstream>  //////////////////////////

/*
 * Write the calibrated h(t) frames to disk.
 *
 * mOutputData must have the channel h already filled (by calling
 * LALComputeStrain at some point before this function is called).
 *
 * The dmt_chans is passed just to get the start time and time step.
 */
int hoftMon::write_frames(const TSeries* dmt_chans[])
{
    if (mIteration < 2)
	return 0;  // we can only calibrate from the third stride on

    /*
     * Extract things from the LAL structures and put it back into
     * something readable by the DMT.
     */
    UINT4 length = mOutputData.h.data->length / 3;  // we want just the middle 1/3

    TSeries hoft;

    hoft.setName("CALIBRATED-H_OF_T");

    // Copy only the middle segment into hoft
    hoft.setData(dmt_chans[0]->getStartTime() - Interval(32),
		 dmt_chans[0]->getTStep(),
		 new DVecType<double>(length,
				      mOutputData.h.data->data + length));
    // The "new" is necessary, because setData destroys the content of
    // the vector that is passed as argument (through the use of a
    // auto_ptr<> that makes a reset())

    // We would still have to write the data back. See Shourov's hMon
    // publishFrames(). It is something like the following:
    mFrameWriter.buildFrame(hoft.getStartTime(), Interval(mStep * 3));
    mFrameWriter.addProcSeries("H2:LSC-STRAIN", hoft);
   // what about this instead???? :
    //mFrameWriter.addChannel("H2:LSC-STRAIN", &&hoft);
    // or (???):
    //  mFrameWriter.addProcSeries("H2:LSC-STRAIN", hoft.extract(
    //		   hoft.getStartTime(), Interval(mStep * 3)));

    mFrameWriter.addHistory("hoftMon", Now() - Interval(32),
			    "real-time time-domain calibrated data");
    mFrameWriter.addWriterHistory();
    mFrameWriter.open("output-hoftmon.gwf", true);
    mFrameWriter.writeFrame();
    mFrameWriter.close();
    // Also, check the documentation on FrWriter at site:ligo.caltech.edu

    /////////////////////////
    if (mIteration == 2) {
	ofstream fout("hoftmon-out-debug.txt");

// 	for (int i = length; i < 2*length; i++)
// 	    fout << mOutputData.h.data->data[i] << endl;
	

	DVector* v = hoft.refDVect();
	double* data = (double*) v->refData();

	for (unsigned int i = 0; i < length; i++)
	    fout << data[i] << endl;

	fout.close();
    }
    /////////////////////////////

    return 0;
}





/* --------------------- Auxiliary functions ---------------------------- */


/*
 * Get pointers to the contents of the different channels.
 */
int hoftMon::get_channels(const TSeries* dmt_chans[])
{
    // Get the channels by calling getDacc()
    for (int i=0; i < 4; i++)
	dmt_chans[i] = getDacc().refData(mChanNames[i]);

    // Check that all channels are there indeed.
    for (int i=0; i < 4; i++)
	if (!(dmt_chans[i]) || (dmt_chans[i])->isEmpty()) {
	    cerr << "Channel " << mChanNames[i] << " was not found.\n";
	    return -1;
	}

    return 0;
}



/* ------------------ Initialization, allocation et al ------------------- */


/*
 * Read command line arguments and initialize internal variables from
 * it. Set all the non-changing input variables (arguments passed to
 * the program). To be called in the constructor.
 */
int hoftMon::initialize_input_data(int argc, const char *argv[])
{
    int err = 0;  /* error code */
    
    /* Initialize default values (fill all but the time series) */

    mInputData.Go.re = -0.44761314397906;  // open loop gain at cal line freq (real)
    mInputData.Go.im = -0.09999284928620;  // (imaginary part)
    mInputData.Do.re = -0.14346135865327;  // digital filter at cal line freq (real)
    mInputData.Do.im =  0.22204243370711;  // (imaginary part)
    mInputData.Wo.re =  0.00987795632736;  // whitening filter at cal line freq (real)
    mInputData.Wo.im = -0.00156482799222;  // (imaginary part)
    mInputData.f     = 407.3;  // frequency of the calibration line
    mInputData.To    = 1.0;    // factors integration time (in s)
    // These default values come from
    // ~xsiemens/hoft/S5RecoveryApril2007/H2/DONEPhase2/S5strainDAGH2_822760000-824862720.ini
    // at ldas-pcdev1.ligo.caltech.edu.
    // They should be given as parameters passed when calling the program.

    mInputData.delta      = 0;
    mInputData.wings      = 16;  // width (in s) of the "extra" data for the calibration
    mInputData.usefactors = 1;
    mInputData.fftconv    = 1;
    mInputData.outalphas  = 0;
    mInputData.darmctrl   = 1;
    mInputData.gamma_fudgefactor = 1;

    /* Read data from the command line. */

    char usage[] =
"Command syntax:\n"
"hoftMon [-help] [-verbose <level>]"
" [-stride <value>]"
" [-olg-re <value>] [-olg-im <value>]"
" [-servo-re <value>] [-servo-im <value>]"
" [-whitener-re <value>] [-whitener-im <value>]"
" [-cal-line-freq <freq>]"
" [-factors-time <time>]"
" [-wings <value>]"
" [-gamma-fudge-factor <value>]"
" [-delta] [-td-fir] [-no-factors] [<DatEnv-args>]\n";

    char help[] =
"Arguments are (a * means that they are not yet implemented):\n"
"  -stride              FLOAT   Length in seconds of the data stride read from the DMT.\n"
"  -cal-line-freq       FLOAT   Calibration line frequency in Hz.\n"
"* -factors-time        FLOAT   Factors integration time in seconds.\n"
"  -olg-re              FLOAT   Real part of the open loop gain at the calibration line frequency.\n"
"  -olg-im              FLOAT   Imaginary part of the open loop gain at the calibration line frequency.\n"
"  -servo-re            FLOAT   Real part of the digital filter at the calibration line frequency.\n"
"  -servo-im            FLOAT   Imaginary part of digital filter at the calibration line frequency.\n"
"  -whitener-re         FLOAT   Real part of the whitening filter at the calibration line frequency.\n"
"  -whitener-im         FLOAT   Imaginary part of whitening filter at the calibration line frequency.\n"
"* -gps-start-time      INT     GPS start time.\n"
"* -gps-end-time        INT     GPS end time.\n"
"  -filters-file        STRING  Name of file containing filters and histories.\n"
"* -frame-cache         STRING  Name of frame cache file.\n"
"* -asq-channel         STRING  AS_Q channel name (eg, L1:LSC-AS_Q).\n"
"* -exc-channel         STRING  Excitation channel name (eg, L1:LSC-ETMX_EXC_DAQ).\n"
"* -darm-channel        STRING  Darm channel name (eg, L1:LSC-DARM_CTRL).\n"
"* -darmerr-channel     STRING  Darm ERR channel name (eg, L1:LSC-DARM_ERR).\n"
"  -wings               INT     Size of wings in seconds.\n"
"* -test-sensing        FLAG    Output residual strain only.\n"
"* -test-actuation      FLAG    Output control strain only.\n"
"  -delta               FLAG    Use unit impulse.\n"
"  -no-factors          FLAG    Do not use factors in strain computation.\n"
"  -td-fir              FLAG    Use time-domain FIR filtering (default is FFT convolution).\n"
"  -gamma-fudge-factor  FLOAT   Fudge factor used to adjust factor values. Gamma is divided by that value.\n"
"* -frame-type          STRING  Frame type to be written (eg, H1_RDS_C01_LX).\n"
"* -strain-channel      STRING  Strain channel name in frame (eg, H1:LSC-STRAIN).\n"
"  -help                FLAG    This message.\n";

    // Look for arguments.
    for (int i = 1; i < argc; i++) {
        string s = argv[i];

	if (s == "-help") {
	    cout << usage << endl << help;
	    finish();  return 0;
	}
	else if (s == "-verbose")       { mVerboseLevel = atoi(argv[++i]); }
	else if (s == "-stride")        { mStep = strtod(argv[++i], 0); }
	else if (s == "-no-factors")    { mInputData.usefactors = 0; }
	else if (s == "-olg-re")        { mInputData.Go.re = strtod(argv[++i], 0); }
	else if (s == "-olg-im")        { mInputData.Go.im = strtod(argv[++i], 0); }
	else if (s == "-servo-re")      { mInputData.Do.re = strtod(argv[++i], 0); }
	else if (s == "-servo-im")      { mInputData.Do.im = strtod(argv[++i], 0); }
	else if (s == "-whitener-re")   { mInputData.Wo.re = strtod(argv[++i], 0); }
	else if (s == "-whitener-im")   { mInputData.Wo.im = strtod(argv[++i], 0); }
	else if (s == "-cal-line-freq") { mInputData.f = strtod(argv[++i], 0); }
	else if (s == "-factors-time")  { mInputData.To = strtod(argv[++i], 0); }
	else if (s == "-delta")         { mInputData.delta = 1; }
	else if (s == "-wings")         { mInputData.wings = atoi(argv[++i]); }
	else if (s == "-td-fir")        { mInputData.fftconv = 0; }
	else if (s == "-gamma-fudge-factor") { mInputData.gamma_fudgefactor = strtod(argv[++i], 0); }
	else if (s == "-filters-file") {
	    err = XLALReadFiltersFile(argv[++i], &mInputData);
	    if (err)
		return err;
 	}
	else if (isDatEnvArg(argv[i])) {
	    i++;
	}
	else {
	    cerr << "Unrecognized argument: " << s << endl
		 << usage;
	    return -3;
	}
    }

    /* Check that the duration makes sense (is long enough) */

    if (mStep < 10.0) {  // actuation filter takes 10 seconds
	cerr << "Time stride is too small (" << mStep << " s) "
	     << "No way I can calibrate this!\n";
	return -1;
    }
    else if (fabs(mStep - 16.0) > 0.01) {  // nothing special about the 0.01
	cerr << "Warning: duration (" << mStep << " s) "
	     << "different from 16 s. I'll continue, though.\n";
    }

//     /* check input data epoch agrees with command line arguments */
//     if ( pInputData->AS_Q.epoch.gpsSeconds != CLA.GPSStart ) {
// 	fprintf(stderr,"GPS start time of data (%d) does not agree with requested start time (%d). Exiting.", 
// 		pInputData->AS_Q.epoch.gpsSeconds, CLA.GPSStart);
// 	return 1;
//     }

    return 0;
}




/*
 * Allocate space looking at how many samples there are in the dmt
 * channels (dmt_chans).
 */
int hoftMon::allocate_data(const TSeries* dmt_chans[])
{
    /* Find the size of the vectors we will need to allocate */

    UINT4 nsamples[4];

    // All samples have the same duration (so we get it from the first)
    REAL8 duration = (REAL8) dmt_chans[0]->getInterval().GetSecs();

    // Number of samples = duration / deltaT
    for (int i=0; i < 4; i++) {
	REAL8 deltaT = (REAL8) dmt_chans[i]->getTStep().GetSecs();
	nsamples[i]  = (UINT4)(duration / deltaT + 0.5);
    }

    // The output data has the same size as AS_Q
    UINT4 nsampl_h = nsamples[0];

    // The alpha et al have a different deltaT
    REAL8 To = mInputData.To;
    UINT4 nsampl_alpha = (UINT4)(duration / To + 0.5);

    /* Check that it makes sense */

    // nsamples as expected from the dmt channels?
    for (int i=0; i < 4; i++) {
	UINT4 n_dmt = (UINT4) dmt_chans[i]->getNSample();
	if (nsamples[i] != n_dmt) {
	    cerr << "Error. Durations differ for channel " << mChanNames[i]
		 << "\nDMT: " << n_dmt << "\tComputed: " << nsamples[i] << endl;
	    return -1;
	}
    }

    /* Allocate space for vectors in LAL input data */

    // The 3* factor accounts for the buffering of data.
    mInputData.AS_Q.data     = XLALCreateREAL4Vector(3 * nsamples[0]);
    mInputData.DARM.data     = XLALCreateREAL4Vector(3 * nsamples[1]);
    mInputData.DARM_ERR.data = XLALCreateREAL4Vector(3 * nsamples[2]);
    mInputData.EXC.data      = XLALCreateREAL4Vector(3 * nsamples[3]);
    if (!mInputData.AS_Q.data || !mInputData.DARM.data ||
	!mInputData.DARM_ERR.data || !mInputData.EXC.data) {
	cerr << "Error allocating space for input channels.\n";
	return -2;
    }

    /* Allocate space for vectors in LAL output data */

    mOutputData.h.data  = XLALCreateREAL8Vector(3 * nsampl_h);
    mOutputData.hC.data = XLALCreateREAL8Vector(3 * nsampl_h);
    mOutputData.hR.data = XLALCreateREAL8Vector(3 * nsampl_h);

    mOutputData.alpha.data     = XLALCreateCOMPLEX16Vector(3 * nsampl_alpha);
    mOutputData.beta.data      = XLALCreateCOMPLEX16Vector(3 * nsampl_alpha);
    mOutputData.alphabeta.data = XLALCreateCOMPLEX16Vector(3 * nsampl_alpha);

    if (!mOutputData.h.data || !mOutputData.hC.data || !mOutputData.hR.data ||
	!mOutputData.alpha.data || !mOutputData.beta.data || !mOutputData.alphabeta.data) {
	cerr << "Error allocating channel vector.\n";
	return -4;
    }
    
    return 0;
}



/*
 * Free the allocated memory.
 */
int hoftMon::free_data()
{
    /* Free input data */
    XLALDestroyREAL4Vector(mInputData.AS_Q.data);
    XLALDestroyREAL4Vector(mInputData.DARM.data);
    XLALDestroyREAL4Vector(mInputData.DARM_ERR.data);
    XLALDestroyREAL4Vector(mInputData.EXC.data);      

    /* Free output data. */
    XLALDestroyREAL8Vector(mOutputData.h.data);
    XLALDestroyREAL8Vector(mOutputData.hC.data);
    XLALDestroyREAL8Vector(mOutputData.hR.data);
    XLALDestroyCOMPLEX16Vector(mOutputData.alpha.data);
    XLALDestroyCOMPLEX16Vector(mOutputData.beta.data);
    XLALDestroyCOMPLEX16Vector(mOutputData.alphabeta.data);

    return 0;
}



/*
 * Destroys the extra objects created at initialize_data(). To be
 * called in the destructor.
 */
int hoftMon::destroy_initialization()
{
    int err = XLALDestroyFiltersFile(&mInputData);
    if (err) {
	cerr << "Error destroying filters file.\n";
	return -1;
    }

    return 0;
}
