/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "TrigStream.hh"
#include "TrigTable.hh"
#include "xsil/ligolw.hh"
#include "xsil/Xwriter.hh"
#include <fstream>
#include <sstream>
#include <unistd.h>
#include <signal.h>

using namespace std;

//======================================  Get a unique ID
static unsigned long mUniqueID(0);

string
getUnique(void) {
    Time idTime(Now());
    idTime.setN(mUniqueID++);
    char line[128];
    TimeStr(idTime, line, "%Y%02m%02d%02H%02N%02S%06n010000");
    return string(line);
}

void
frog(const string& s) {
    sigignore(SIGCLD);
    if (fork()) return;
    string a(s);
    a += char(0);
    const char* argv[32];
    argv[0] = 0;
    char* p = &a[0];
    for (int i=0; i<31 && *p; ) {
        while (*p == ' ') *p++ = 0;
	if (*p) argv[i] = p;
	while (*p && *p != ' ') p++;
	argv[++i] = 0;
    }
    exit(execvp(argv[0], const_cast<char* const*>(argv)));
}

//======================================  Parameter constructor
TrigStream::param::param(void)
  : _Inactive(300.0), _MaxDelay(300.0), _TableCount(defTableCount)
{}

//======================================  Statistics constructor
TrigStream::stream_stats::stream_stats(void)
  : errTot(0), fileTot(0), segTot(0), trigTot(0)
{
    lastUpdate = Now();
    lastWrite  = lastUpdate;
}

//======================================  Count a segment
void
TrigStream::stream_stats::count_seg(void) {
    segTot++;
    lastUpdate = Now();
}

//======================================  Count a trigger
void
TrigStream::stream_stats::count_trig(void) {
    trigTot++;
    lastUpdate = Now();
}

//======================================  Count a trigger
void
TrigStream::stream_stats::count_write(const string& file) {
    fileTot++;
    lastWrite = Now();
    lastFile  = file;
}

//======================================  Stream constructor
TrigStream::TrigStream(const std::string& name, mask_type mask)
  : mName(name), mDebug(0), mDestMsk(mask), mExcludeMsk(0)
{
}

TrigStream::~TrigStream(void) {
}

//======================================  Add a process
void 
TrigStream::addProcess(const Process& pdb) {
    mProcDB[pdb.refData().getProcessID()] = pdb;
}

//======================================  Add a segment
bool
TrigStream::addSegment(const trig::Segment& seg, mask_type enable) {
   return true;
}

//======================================  Add a Trigger 
bool
TrigStream::addTrigger(const trig::TrigBase& trg, mask_type enable) {
    return true;
}

//=======================================  Get trigger and segment flags
bool
TrigStream::empty(void) const {
    return !getNSegs() && !getNTrigs();
}

TrigStream::count_type
TrigStream::getNSegs(void) const {
    return 0;
}

TrigStream::count_type
TrigStream::getNTrigs(void) const {
    return 0;
}

//======================================  Get estimated timeout
Interval 
TrigStream::getTimeOut(const Time& now) const {
    Interval dTWrite  = mStats.lastWrite  + mParam._MaxDelay - now;
    if (double(dTWrite) < 0.0) dTWrite = 0.0;
    Interval dTUpdate = mStats.lastUpdate + mParam._Inactive - now;
    if (double(dTUpdate) < 0.0) dTUpdate = 0.0;
    if (dTWrite <= dTUpdate) return dTWrite;
    return dTUpdate;
}

//======================================  Get estimated timeout
void 
TrigStream::PurgeProcDB(void) {
}

//======================================  Get estimated timeout
void 
TrigStream::removeProc(const std::string& ident) {
    if (mProcDB.find(ident) == mProcDB.end()) return;
    mProcDB[ident].setClosed();
    if (!mProcDB[ident].isReleasable()) mProcDB.erase(ident);  
}

//======================================  Set parameters
void 
TrigStream::setDebug(int lvl) {
    mDebug=lvl;
}

void 
TrigStream::setDestMsk(mask_type mask) {
    mDestMsk = mask;
}

void 
TrigStream::setExcludeMsk(mask_type mask) {
    mExcludeMsk = mask;
}

void 
TrigStream::setInactive(Interval dt) {
    mParam._Inactive = dt;
}

void 
TrigStream::setMaxDelay(Interval dt) {
    mParam._MaxDelay = dt;
}

void 
TrigStream::setOutDir(const std::string& dir) {
    mParam._OutDir = dir;
    string::size_type l = mParam._OutDir.size();
    if (l > 0 && mParam._OutDir[l-1] != '/') mParam._OutDir += "/";
}

void 
TrigStream::setParam(const param& par) {
    mParam = par;
}

//======================================  Test destination flags
bool 
TrigStream::testWrite(const trig::Segment& seg, mask_type dest) const {
    return (dest & mDestMsk) != 0 && (dest & mExcludeMsk) == 0;
}

bool 
TrigStream::testWrite(const trig::TrigBase& tr, mask_type dest) const {
    mask_type msgd = tr.getDisposition();
    return (dest & mDestMsk & msgd) != 0 && (dest & mExcludeMsk & msgd) == 0;
}

//======================================  See if its time to write
bool
TrigStream::writeNow(bool force) const {
    if (empty()) return false;
    Time now(Now());
    return (force || 
	    getNTrigs() >= mParam._TableCount || 
	    getTimeOut(now) <= Interval(0.0));
}

//======================================  Write DB
void 
TrigStream::WriteDB(bool force) {
}

//======================================  Update
void
TrigStream::tick(const Time& now) {
    if (empty()) {
	mStats.lastUpdate = now;
    }
}

//======================================  Epics stream class
EpxStream::EpxStream(const std::string& name, mask_type mask)
  : TrigStream(name, mask)
{}

//======================================  Epics stream destructor
EpxStream::~EpxStream(void) {
}

//======================================  Clone an Epics stream
EpxStream*
EpxStream::clone(void) const {
    return new EpxStream(*this);
}

//======================================  Send the trigger to Epics
bool
EpxStream::addTrigger(const trig::TrigBase& trg, mask_type enable) {
    if (!testWrite(trg, enable)) return false;
    string procid = trg.getProcess();
    proc_iter   iProc  = mProcDB.find(procid);
    if (iProc == mProcDB.end()) return false;
 
    stringstream ingest;
    ingest << "epics_alarm ";      // submission command
    switch (trg.getPriority()) {
    case trig::p_info:
        ingest << "Information ";
	break;
    case trig::p_warn:
        ingest << "Warning ";
	break;
    case trig::p_error:
        ingest << "Error ";
	break;
    case trig::p_severe:
        ingest << "Severe ";
	break;
    }
    ingest << trg.getTime().getS() << " \"";
    ingest << iProc->second.refData().getName() << " ";
    ingest << trg.getID() << " " << trg.getSubID();
    ingest << "\" > /dev/null";
    if (getDebug()) cout << "Epics Alarm: " << ingest.str() << endl;
    frog(ingest.str());
    mStats.count_trig();
    return true;
}

EpxStream::count_type
EpxStream::getNSegs(void) const {
    return 0;
}

EpxStream::count_type
EpxStream::getNTrigs(void) const {
    return 0;
}

//======================================  Ldas DB stream class
LdasStream::LdasStream(const std::string& name, mask_type mask)
  : TrigStream(name, mask)
{}

//======================================  Ldas DB stream destructor
LdasStream::~LdasStream(void) {
}

//======================================  Clone an Ldas stream
LdasStream*
LdasStream::clone(void) const {
    return new LdasStream(*this);
}

//======================================  Add a segment
bool
LdasStream::addSegment(const trig::Segment& seg, mask_type enable) {
    if (!testWrite(seg, enable)) return false;
    std::string procid = seg.getProcess();
    proc_iter   iProc  = mProcDB.find(procid);
    if (iProc == mProcDB.end()) return false;
    iProc->second.incrementWrite();
    trig::Segment s(seg);
    string grpName(s.getGroup());
    if (*s.getIfos() != 0 && grpName.find(":") == string::npos) {
	grpName.insert(0,":");
	grpName.insert(0,s.getIfos());
	s.setGroup(grpName.c_str());
    }
    mWriter.addSegment(s, iProc->second.refData());
    mStats.count_seg();
    return true;
}

//======================================  Add a Trigger 
bool
LdasStream::addTrigger(const trig::TrigBase& trg, mask_type enable) {
    if (!testWrite(trg, enable)) return false;
    std::string procid = trg.getProcess();
    proc_iter   iProc  = mProcDB.find(procid);
    if (iProc == mProcDB.end()) return false;
    iProc->second.incrementWrite();
    mWriter.addTrigger(trg, iProc->second.refData());
    mStats.count_trig();
    return true;
}

//======================================  Data size methods
bool
LdasStream::empty(void) const {
    return mWriter.empty();
}

LdasStream::count_type
LdasStream::getNSegs(void) const {
    return mWriter.getNSegs();
}

LdasStream::count_type
LdasStream::getNTrigs(void) const {
    return mWriter.getNTrigs();
}

//======================================  Write DB
void 
LdasStream::WriteDB(bool force) {
    if ( !writeNow(force) ) return;

    //----------------------------------  Write the trigger xml file
    string outFile("TrigMgr");
    outFile += getUnique() + ".xml";
    string path(mParam._OutDir + outFile);
    if (getDebug() > 1) cout << "Stream: " << getName() << " writing file: " 
			     << path << endl;
    mWriter.write(path);
    mStats.count_write(path);

    //----------------------------------  Submit request to the metadatabase.
    string ingest("SeqInsert ");  // SeqInsert command
    if (getenv("DMTCOMNODE")) {   // Node name (if DMTCOMNODE specified)
        ingest += "-host ";
	ingest += getenv("DMTCOMNODE");
	ingest += " ";
    }
    ingest += "submit ";          // Submit command
    ingest += outFile;            // xml file name
    frog(ingest);

    //----------------------------------  Check pointers are all valid.
    mWriter.clear();
}

//======================================  Segment DB stream class
SegDBStream::SegDBStream(const string& name, mask_type mask, const string& dbname)
  : TrigStream(name, mask)
{}

//======================================  Segment DB stream destructor
SegDBStream::~SegDBStream(void) {
}

//======================================  Clone an Segment DB stream
SegDBStream*
SegDBStream::clone(void) const {
    return new SegDBStream(*this);
}

//======================================  Add a segment
bool
SegDBStream::addSegment(const trig::Segment& seg, mask_type enable) {
    if (!testWrite(seg, enable)) return false;
    std::string procid = seg.getProcess();
    proc_iter   iProc  = mProcDB.find(procid);
    if (iProc == mProcDB.end()) return false;
    iProc->second.incrementWrite();
    mWriter.addSegment(seg, iProc->second.refData());
    mStats.count_seg();
    return true;
}

//======================================  Data size methods
bool
SegDBStream::empty(void) const {
    return mWriter.empty();
}

SegDBStream::count_type
SegDBStream::getNSegs(void) const {
    return mWriter.getNSegs();
}

SegDBStream::count_type
SegDBStream::getNTrigs(void) const {
    return mWriter.getNTrigs();
}

//======================================  Write DB
void 
SegDBStream::WriteDB(bool force) {
    if ( !writeNow(force) ) return;

    //----------------------------------  Write the trigger xml file
    string outFile("SegmDB");
    outFile += getUnique() + ".xml";
    string path(mParam._OutDir + outFile);
    if (getDebug() > 1) cout << "Stream: " << getName() << " writing file: " 
			     << path << endl;
    mWriter.write(path);
    mStats.count_write(path);

    //----------------------------------  Submit to the segment data base.
    string ingest("dmtdq_seg_insert");  // Segment insert command
    if (getenv("DMTSEGSERVER")) {   // Node name (if DMTSEGSERVER specified)
        ingest += " --server ";
	ingest += getenv("DMTSEGSERVER");
    }
    ingest += " --file ";          // Submit command
    ingest += outFile;            // xml file name
    frog(ingest);
    //cout << "DQSegDB: File " << outFile << " written" << endl;

    //----------------------------------  Check pointers are all valid.
    mWriter.clear();
}
