/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef TRIGPAR_HH
#define TRIGPAR_HH
//
//    TrigPar  passes the parameters for a trigger filter or detector.
//
//    include templates
#include <map>
#include <functional>
#include <string>
#include <iosfwd>

class Param;

/**  A trigger parameter dictionary holds a list of parameters and their 
  *  values. Each parameter can be of type int, double, or string.
  *  @memo Trigger parameter dictionary class.
  *  @author John G Zweizig
  *  @version 1.2; Last Modified April 15, 1999.
  */
class TrigPar {
public:
    /**  Parameter dictionary default constructor.
      *  @memo Parameter dictionary default constructor.
      */
    TrigPar(void) {}

    /**  Make a dictionary and add parameters as described by the file.
      *  @memo Initializing constructor.
      */
    TrigPar(const char *FileName);

    /**  Make a dictionary and add parameters as described by the file.
      *  @memo Initializing constructor.
      */
    TrigPar(std::istream& file);

    /** Dictionary destructor.
      * @memo Dictionary destructor.
      */
  ~TrigPar(void);

    /**  Test whether the named parameter is defined.
      *  @memo Test whether the named parameter is defined.
      */
    bool exists(const std::string& ParName) const;

    /**  The named parameter may be assigned a value (e.g. Dict["MyPar"] = 5;)
      *  or accessed (e.g. int = int(Dict["MyPar"]);) using the [] operator.
      *  If an undefined name is used on the left-hand-side of an assignment,
      *  the specified parameter is created with the type of the rhs of the
      *  assignment.
      *  @memo Reference to the specified parameter block.
      */
    Param& operator[](const std::string &Name);

    /**  The named parameter may be assigned a value (e.g. Dict["MyPar"] = 5;)
      *  or accessed (e.g. int = int(Dict["MyPar"]);) using the [] operator.
      *  Referencing an undefined parameter will cause an exception.
      *  @memo Constant reference to a specified parameter block.
      *  @return constant reference to the requested parameter object.
      */
    const Param& operator[](const std::string &Name) const;

    /**  The value of the specified parameter is converted to integer if 
      *  necessary and returned.
      *  @memo Get an integer parameter value.
      */
    int getInt(const std::string& ParName) const;

    /**  Get the parameter value (converted to double).
      */
    double getDouble(const std::string& ParName) const;

    /**  Get the parameter value (converted to string).
      */
    std::string getString(const std::string& ParName) const;

    /**  Create an integer parameter and add it to the dictionary.
      *  @memo Add an integer parameter.
      */
    void addPar(const std::string& ParName, int ipar);

    /**  Add a double (float) parameter.
      *  @memo Add a double (float) parameter.
      */
    void addPar(const std::string& ParName, double dpar);

    /**  Add a string parameter.
      *  @memo Add a string parameter.
      */
    void addPar(const std::string& ParName, const std::string& spar);

    /**  Add a parameter of the specified type.
      *  @memo Add a parameter of the specified type.
      */
    void addPar(const std::string& Name, const char *Type, const char *Value);

    /**  Remove the specified parameter.
      *  @memo Remove the specified parameter.
      */
    void remove(const std::string& Name);

    /**  Set parameter to an int
      *  @memo Set parameter to an int
      */
    void setPar(const std::string& ParName, int ipar);

    /**  Set parameter to a double
      *  @memo Set parameter to a double
      */
    void setPar(const std::string& ParName, double dpar);

    /**  Set parameter to a string
      *  @memo Set parameter to a string
      */
    void setPar(const std::string& ParName, const std::string& spar);

    /**  Fill parameter dictionary from the specified file.
      *  @memo Fill parameter dictionary from the specified file.
      */
    void Read(std::istream& istr, const std::string& prefix="");

    /**  Write the parameters, their types and values to a stream.
      *  @memo Write the parameters, their types and values to a stream.
      */
    std::ostream& Write(std::ostream& ostr, const std::string& prefix="") const;

  private:
    /**  Compare string names.
      */
    typedef std::less<std::string> scmp;
    typedef std::map<std::string, Param*, scmp> ParMap;
    typedef ParMap::iterator map_iter;
    typedef ParMap::const_iterator const_map_iter;
    ParMap  mParSet;
};

//--------------------------------------  Inline functions.
inline Param& 
TrigPar::operator[](const std::string &Name) {
    return *mParSet[Name];}

inline const Param& 
TrigPar::operator[](const std::string &Name) const {
    return *(mParSet.find(Name)->second);}


#endif  //  TRIGPAR_HH
